﻿using System;
using System.Collections.Generic;
using System.Linq;

using Achiral;
using Achiral.Extension;
using System.Linq.Expressions;
using System.Threading;


public interface IAccumulable<TSource, TResult>
{
    IAccumulator<TSource, TResult> GetAccumulator();
}
public interface IAccumulator<TSource, TResult> : IDisposable
{
    bool Process(TSource item);
    TResult Result { get; }
}

public static partial class Accumulable
{
    public static TResult Accumulate<TSource, TResult>(this IEnumerable<TSource> source, IAccumulable<TSource, TResult> accumulable)
    {
        using (var accumulator = accumulable.GetAccumulator())
        {
            foreach (var item in source)
            {
                if (!accumulator.Process(item)) break;
            }
            return accumulator.Result;
        }
    }

    public static IAccumulable<TSource, List<TSource>> ToList<TSource>()
    {
        return new Aggregatable<TSource, List<TSource>>(new List<TSource>(), (list, item) => { list.Add(item); return list; });
    }
    public static IAccumulable<TSource, TSource[]> ToArray<TSource>()
    {
        return new Aggregatable<TSource, List<TSource>, TSource[]>(new List<TSource>(), (list, item) => { list.Add(item); return list; }, list => list.ToArray());
    }
    public static IAccumulable<TSource, int> Count<TSource>()
    {
        return new Aggregatable<TSource, int>(0, (count, _) => count + 1);
    }
    public static IAccumulable<TSource, TSource> Min<TSource>()
    {
        return new AggregatableSelfSeed<TSource>((min, x) => Comparer<TSource>.Default.Compare(min, x) < 0 ? min : x);
    }
    public static IAccumulable<TSource, TSource> Max<TSource>()
    {
        return new AggregatableSelfSeed<TSource>((max, x) => Comparer<TSource>.Default.Compare(max, x) > 0 ? max : x);
    }
    private static class Operators<T>
    {
        public static Func<T, T, T> Add { get { return Operators<T, T, T>.Add; } }
        public static Func<T, T, T> Subtract { get { return Operators<T, T, T>.Subtract; } }
        public static Func<T, T, T> Multiply { get { return Operators<T, T, T>.Multiply; } }
        public static Func<T, T, T> Divide { get { return Operators<T, T, T>.Divide; } }
    }
    private static class Converters<TSource, TResult>
    {
        private volatile static Func<TSource, TResult> _conv;
        public static Func<TSource, TResult> Default
        {
            get
            {
                while (_conv == null)
                {
                    var a = Expression.Parameter(typeof(TSource), "source");
                    var f = Expression.Lambda<Func<TSource, TResult>>(Expression.Convert(a, typeof(TResult)), a).Compile();
                    Interlocked.CompareExchange(ref _conv, f, null);
                }
                return _conv;
            }
        }
    }
    private static class Operators<T1, T2, TResult>
    {
        private volatile static Func<T1, T2, TResult> _add;
        private volatile static Func<T1, T2, TResult> _sub;
        private volatile static Func<T1, T2, TResult> _mul;
        private volatile static Func<T1, T2, TResult> _div;
        public static Func<T1, T2, TResult> Add
        {
            get
            {
                while (_add == null)
                {
                    var a = Expression.Parameter(typeof(T1), "a");
                    var b = Expression.Parameter(typeof(T2), "b");
                    var f = Expression.Lambda<Func<T1, T2, TResult>>(Expression.Add(a, b), a, b).Compile();
                    Interlocked.CompareExchange(ref _add, f, null);
                }
                return _add;
            }
        }
        public static Func<T1, T2, TResult> Subtract
        {
            get
            {
                while (_sub == null)
                {
                    var a = Expression.Parameter(typeof(T1), "a");
                    var b = Expression.Parameter(typeof(T2), "b");
                    var f = Expression.Lambda<Func<T1, T2, TResult>>(Expression.Subtract(a, b), a, b).Compile();
                    Interlocked.CompareExchange(ref _sub, f, null);
                }
                return _sub;
            }
        }
        public static Func<T1, T2, TResult> Multiply
        {
            get
            {
                while (_mul == null)
                {
                    var a = Expression.Parameter(typeof(T1), "a");
                    var b = Expression.Parameter(typeof(T2), "b");
                    var f = Expression.Lambda<Func<T1, T2, TResult>>(Expression.Multiply(a, b), a, b).Compile();
                    Interlocked.CompareExchange(ref _mul, f, null);
                }
                return _mul;
            }
        }
        public static Func<T1, T2, TResult> Divide
        {
            get
            {
                while (_div == null)
                {
                    var a = Expression.Parameter(typeof(T1), "a");
                    var b = Expression.Parameter(typeof(T2), "b");
                    var f = Expression.Lambda<Func<T1, T2, TResult>>(Expression.Divide(a, b), a, b).Compile();
                    Interlocked.CompareExchange(ref _div, f, null);
                }
                return _div;
            }
        }
    }

    public static IAccumulable<TSource, TResult> Average<TSource, TResult>()
    {
        return Select<TSource, TResult, TResult>(
            Converters<TSource, TResult>.Default,
            new AggregatableSelfSeed<TResult, Tuple<TResult, int>, TResult>
                (
                    (pair, x) => Make.Tuple(Operators<TResult>.Add(pair.Arg1, x), pair.Arg2 + 1),
                    x => Make.Tuple(x, 1),
                    pair => Operators<TResult, TResult, TResult>.Divide(pair.Arg1, Converters<int, TResult>.Default(pair.Arg2))
                )
        );
    }
    public static IAccumulable<TSource, TSource> Sum<TSource>()
    {
        return new Aggregatable<TSource, TSource>(default(TSource), Operators<TSource>.Add);
    }
    public static IAccumulable<TSource, TResult> Where<TSource, TResult>(Func<TSource, bool> predicate, IAccumulable<TSource, TResult> accumulable)
    {
        return new WhereImpl<TSource, TResult>(predicate, accumulable);
    }
    public static IAccumulable<TSource, TResult> Select<TSource, TProjected, TResult>(Func<TSource, TProjected> selector, IAccumulable<TProjected, TResult> accumulable)
    {
        return new SelectImpl<TSource, TProjected, TResult>(selector, accumulable);
    }
}

public static partial class Accumulable
{
    private sealed class Aggregatable<TSource, TAccumulate, TResult> : IAccumulable<TSource, TResult>
    {
        private TAccumulate _seed;
        private Func<TAccumulate, TSource, TAccumulate> _func;
        private Func<TAccumulate, TSource, bool> _conditionSelector;
        private Func<TAccumulate, TResult> _resultSelector;
        private Action<TAccumulate> _disposer;

        public Aggregatable(TAccumulate seed, Func<TAccumulate, TSource, TAccumulate> func, Func<TAccumulate, TSource, bool> conditionSelector, Func<TAccumulate, TResult> resultSelector, Action<TAccumulate> disposer)
        {
            this._seed = seed;
            this._func = func;
            this._conditionSelector = conditionSelector;
            this._resultSelector = resultSelector;
            this._disposer = disposer;
        }
        public Aggregatable(TAccumulate seed, Func<TAccumulate, TSource, TAccumulate> func, Func<TAccumulate, TResult> resultSelector)
            : this(seed, func, (_, __) => true, resultSelector, _ => { })
        {
        }
        public IAccumulator<TSource, TResult> GetAccumulator()
        {
            return new Aggregator(_seed, _func, _conditionSelector, _resultSelector, _disposer);
        }
        private sealed class Aggregator : IAccumulator<TSource, TResult>
        {
            private TAccumulate _accumulator;
            private readonly Func<TAccumulate, TSource, TAccumulate> _func;
            private readonly Func<TAccumulate, TSource, bool> _conditionSelector;
            private readonly Func<TAccumulate, TResult> _resultSelector;
            private readonly Action<TAccumulate> _disposer;

            public Aggregator(TAccumulate seed, Func<TAccumulate, TSource, TAccumulate> func, Func<TAccumulate, TSource, bool> conditionSelector, Func<TAccumulate, TResult> resultSelector, Action<TAccumulate> disposer)
            {
                this._accumulator = seed;
                this._func = func;
                this._conditionSelector = conditionSelector;
                this._resultSelector = resultSelector;
                this._disposer = disposer;
            }

            public bool Process(TSource item)
            {
                _accumulator = _func(_accumulator, item);
                return _conditionSelector(_accumulator, item);
            }
            public TResult Result
            {
                get { return _resultSelector(_accumulator); }
            }
            public void Dispose()
            {
                _disposer(_accumulator);
            }
        }
    }

    private sealed class AggregatableSelfSeed<TSource> : IAccumulable<TSource, TSource>
    {
        private readonly Func<TSource, TSource, TSource> _func;
        private readonly Func<TSource, TSource, bool> _conditionSelector;
        private readonly Func<TSource, TSource> _resultSelector;
        private readonly Action<TSource> _disposer;

        public AggregatableSelfSeed(Func<TSource, TSource, TSource> func, Func<TSource, TSource, bool> conditionSelector, Func<TSource, TSource> resultSelector, Action<TSource> disposer)
        {
            this._func = func;
            this._conditionSelector = conditionSelector;
            this._resultSelector = resultSelector;
            this._disposer = disposer;
        }
        public AggregatableSelfSeed(Func<TSource, TSource, TSource> func)
            : this(func, (_, __) => true, result => result, _ => { })
        {
        }
        public AggregatableSelfSeed(Func<TSource, TSource, TSource> func, Func<TSource, TSource> resultSelector)
            : this(func, (_, __) => true, resultSelector, _ => { })
        {
        }

        public IAccumulator<TSource, TSource> GetAccumulator()
        {
            return new Aggregator(_func, _conditionSelector, _resultSelector, _disposer);
        }
        private sealed class Aggregator : IAccumulator<TSource, TSource>
        {
            private TSource _accumulator;
            private bool _isFirst;
            private readonly Func<TSource, TSource, TSource> _func;
            private readonly Func<TSource, TSource, bool> _conditionSelector;
            private readonly Func<TSource, TSource> _resultSelector;
            private readonly Action<TSource> _disposer;

            public Aggregator(Func<TSource, TSource, TSource> func, Func<TSource, TSource, bool> conditionSelector, Func<TSource, TSource> resultSelector, Action<TSource> disposer)
            {
                this._accumulator = default(TSource);
                this._isFirst = true;
                this._func = func;
                this._conditionSelector = conditionSelector;
                this._resultSelector = resultSelector;
                this._disposer = disposer;
            }
            public bool Process(TSource item)
            {
                if (_isFirst)
                {
                    _accumulator = item;
                    _isFirst = false;
                }
                else
                {
                    _accumulator = _func(_accumulator, item);
                }
                return _conditionSelector(_accumulator, item);
            }
            public TSource Result
            {
                get
                {
                    if (_isFirst) throw new InvalidOperationException("No element has processed");
                    return _resultSelector(_accumulator);
                }
            }
            public void Dispose()
            {
                _disposer(_accumulator);
            }
        }
    }

    private sealed class AggregatableSelfSeed<TSource, TAccumulate, TResult> : IAccumulable<TSource, TResult>
    {
        private readonly Func<TAccumulate, TSource, TAccumulate> _func;
        private readonly Func<TSource, TAccumulate> _seedSelector;
        private readonly Func<TAccumulate, TSource, bool> _conditionSelector;
        private readonly Func<TAccumulate, TResult> _resultSelector;
        private readonly Action<TAccumulate> _disposer;

        public AggregatableSelfSeed(Func<TAccumulate, TSource, TAccumulate> func, Func<TSource, TAccumulate> seedSelector, Func<TAccumulate, TSource, bool> conditionSelector, Func<TAccumulate, TResult> resultSelector, Action<TAccumulate> disposer)
        {
            this._func = func;
            this._seedSelector = seedSelector;
            this._conditionSelector = conditionSelector;
            this._resultSelector = resultSelector;
            this._disposer = disposer;
        }
        public AggregatableSelfSeed(Func<TAccumulate, TSource, TAccumulate> func, Func<TSource, TAccumulate> seedSelector, Func<TAccumulate, TResult> resultSelector)
            : this(func, seedSelector, (_, __) => true, resultSelector, _ => { })
        {
        }
        public IAccumulator<TSource, TResult> GetAccumulator()
        {
            return new Aggregator(_func, _seedSelector, _conditionSelector, _resultSelector, _disposer);
        }
        private sealed class Aggregator : IAccumulator<TSource, TResult>
        {
            private TAccumulate _accumulator;
            private bool _isFirst;
            private readonly Func<TAccumulate, TSource, TAccumulate> _func;
            private readonly Func<TSource, TAccumulate> _seedSelector;
            private readonly Func<TAccumulate, TSource, bool> _conditionSelector;
            private readonly Func<TAccumulate, TResult> _resultSelector;
            private readonly Action<TAccumulate> _disposer;

            public Aggregator(Func<TAccumulate, TSource, TAccumulate> func, Func<TSource, TAccumulate> seedSelector, Func<TAccumulate, TSource, bool> conditionSelector, Func<TAccumulate, TResult> resultSelector, Action<TAccumulate> disposer)
            {
                this._accumulator = default(TAccumulate);
                this._isFirst = true;
                this._func = func;
                this._seedSelector = seedSelector;
                this._conditionSelector = conditionSelector;
                this._resultSelector = resultSelector;
                this._disposer = disposer;
            }
            public bool Process(TSource item)
            {
                if (_isFirst)
                {
                    _accumulator = _seedSelector(item);
                    _isFirst = false;
                }
                else
                {
                    _accumulator = _func(_accumulator, item);
                }
                return _conditionSelector(_accumulator, item);
            }
            public TResult Result
            {
                get
                {
                    if (_isFirst) throw new InvalidOperationException("No element has processed");
                    return _resultSelector(_accumulator);
                }
            }
            public void Dispose()
            {
                _disposer(_accumulator);
            }
        }
    }

    private sealed class Aggregatable<TSource, TResult> : IAccumulable<TSource, TResult>
    {
        private TResult _seed;
        private readonly Func<TResult, TSource, TResult> _func;
        private readonly Func<TResult, TSource, bool> _conditionSelector;
        private readonly Func<TResult, TResult> _resultSelector;
        private readonly Action<TResult> _disposer;

        public Aggregatable(TResult seed, Func<TResult, TSource, TResult> func, Func<TResult, TSource, bool> conditionSelector, Func<TResult, TResult> resultSelector, Action<TResult> disposer)
        {
            this._seed = seed;
            this._func = func;
            this._conditionSelector = conditionSelector;
            this._resultSelector = resultSelector;
            this._disposer = disposer;
        }
        public Aggregatable(TResult seed, Func<TResult, TSource, TResult> func, Func<TResult, TResult> resultSelector)
            : this(seed, func, (_, __) => true, resultSelector, _ => { })
        {
        }
        public Aggregatable(TResult seed, Func<TResult, TSource, TResult> func)
            : this(seed, func, (_, __) => true, result => result, _ => { })
        {
        }

        public IAccumulator<TSource, TResult> GetAccumulator()
        {
            return new Aggregator(_seed, _func, _conditionSelector, _resultSelector, _disposer);
        }
        private sealed class Aggregator : IAccumulator<TSource, TResult>
        {
            private TResult _accumulator;
            private Func<TResult, TSource, TResult> _func;
            private Func<TResult, TSource, bool> _conditionSelector;
            private Func<TResult, TResult> _resultSelector;
            private Action<TResult> _disposer;

            public Aggregator(TResult seed, Func<TResult, TSource, TResult> func, Func<TResult, TSource, bool> conditionSelector, Func<TResult, TResult> resultSelector, Action<TResult> disposer)
            {
                this._accumulator = seed;
                this._func = func;
                this._conditionSelector = conditionSelector;
                this._resultSelector = resultSelector;
                this._disposer = disposer;
            }
            public bool Process(TSource item)
            {
                _accumulator = _func(_accumulator, item);
                return _conditionSelector(_accumulator, item);
            }
            public TResult Result
            {
                get { return _resultSelector(_accumulator); }
            }
            public void Dispose()
            {
                _disposer(_accumulator);
            }
        }
    }

    internal class AccumulatorBase<TSource, TResult> : IAccumulator<TSource, TResult>
    {
        protected readonly IAccumulable<TSource, TResult> BaseAccumulable;
        protected IAccumulator<TSource, TResult> RawBaseAccumulator = null;
        protected IAccumulator<TSource, TResult> SafeBaseAccumulator
        {
            get
            {
                if (RawBaseAccumulator == null)
                {
                    RawBaseAccumulator = BaseAccumulable.GetAccumulator();
                }
                return RawBaseAccumulator;
            }
        }
        internal AccumulatorBase(IAccumulable<TSource, TResult> accumulable)
        {
            this.BaseAccumulable = accumulable;
        }
        public virtual bool Process(TSource item)
        {
            return SafeBaseAccumulator.Process(item);
        }
        public virtual TResult Result
        {
            get
            {
                return SafeBaseAccumulator.Result;
            }
        }
        public void Dispose()
        {
            if (RawBaseAccumulator != null)
            {
                RawBaseAccumulator.Dispose();
            }
        }
    }

    internal abstract class AccumulatorBase<TSource, TProjected, TResult> : IAccumulator<TSource, TResult>
    {
        protected readonly IAccumulable<TProjected, TResult> BaseAccumulable;
        protected IAccumulator<TProjected, TResult> RawBaseAccumulator = null;
        protected IAccumulator<TProjected, TResult> SafeBaseAccumulator
        {
            get
            {
                if (RawBaseAccumulator == null)
                {
                    RawBaseAccumulator = BaseAccumulable.GetAccumulator();
                }
                return RawBaseAccumulator;
            }
        }
        internal AccumulatorBase(IAccumulable<TProjected, TResult> accumulable)
        {
            this.BaseAccumulable = accumulable;
        }
        public abstract bool Process(TSource item);
        protected bool Process(TProjected item)
        {
            return this.SafeBaseAccumulator.Process(item);
        }
        public virtual TResult Result
        {
            get
            {
                return SafeBaseAccumulator.Result;
            }
        }
        public void Dispose()
        {
            if (RawBaseAccumulator != null)
            {
                RawBaseAccumulator.Dispose();
            }
        }
    }
}

public static partial class Accumulable
{
    private sealed class WhereImpl<TSource, TResult> : IAccumulable<TSource, TResult>
    {
        private readonly Func<TSource, bool> _predicate;
        private readonly IAccumulable<TSource, TResult> _accumulatable;
        internal WhereImpl(Func<TSource, bool> predicate, IAccumulable<TSource, TResult> accumulable)
        {
            this._predicate = predicate;
            this._accumulatable = accumulable;
        }
        public IAccumulator<TSource, TResult> GetAccumulator()
        {
            return new Accumulator(_predicate, _accumulatable);
        }
        private sealed class Accumulator : AccumulatorBase<TSource, TResult>
        {
            private readonly Func<TSource, bool> _predicate;
            internal Accumulator(Func<TSource, bool> predicate, IAccumulable<TSource, TResult> accumulable)
                : base(accumulable)
            {
                this._predicate = predicate;
            }
            public override bool Process(TSource item)
            {
                if (!_predicate(item)) return true;
                return base.Process(item);
            }
        }
    }
    private sealed class SelectImpl<TSource, TProjected, TResult> : IAccumulable<TSource, TResult>
    {
        private readonly Func<TSource, TProjected> _selector;
        private readonly IAccumulable<TProjected, TResult> _accumulatable;
        internal SelectImpl(Func<TSource, TProjected> selector, IAccumulable<TProjected, TResult> accumulable)
        {
            this._selector = selector;
            this._accumulatable = accumulable;
        }
        public IAccumulator<TSource, TResult> GetAccumulator()
        {
            return new Accumulator(_selector, _accumulatable);
        }
        private sealed class Accumulator : AccumulatorBase<TSource, TProjected, TResult>
        {
            private readonly Func<TSource, TProjected> _selector;
            internal Accumulator(Func<TSource, TProjected> selector, IAccumulable<TProjected, TResult> accumulable)
                : base(accumulable)
            {
                this._selector = selector;
            }
            public override bool Process(TSource item)
            {
                return base.Process(_selector(item));
            }
        }
    }
}