﻿using System;
using System.Collections.Generic;
using System.Linq;
using Achiral.Extension;
using System.Linq.Expressions;
using System.Diagnostics;

public static partial class Accumulable
{
    public static TResult ConcurrentReduce<TResult>(
        this IEnumerable<int> source,
        Expression<Func<IEnumerable<int>, TResult>> expr)
    {
        var seq = expr.Parameters[0];

        var newexpr = expr.Body as NewExpression;

        var accumulators = newexpr.Arguments
                                  .OfType<MethodCallExpression>()
                                  .Select(subexpr => ConvertLeafToRoot(subexpr))
                                  .Select(accumulable => accumulable.GetAccumulator())
                                  .ToArray();

        foreach (var item in source)
        {
            foreach (var accumulator in accumulators)
            {
                accumulator.Process(item);
            }
        }

        var results = accumulators.Select(accumulator => accumulator.Result)
                                  .Cast<object>()
                                  .ToArray();

        foreach (var accumulator in accumulators)
        {
            accumulator.Dispose();
        }

        return (TResult)newexpr.Constructor.Invoke(results);
    }
    private static IAccumulable<int, int> ConvertLeafToRoot(MethodCallExpression call)
    {
        var accumulable = default(IAccumulable<int, int>);
        switch (call.Method.Name)
        {
            case "Sum":
                accumulable = Accumulable.Sum<int>();
                break;
            case "Count":
                accumulable = Accumulable.Count<int>();
                break;
            case "Min":
                accumulable = Accumulable.Min<int>();
                break;
            case "Max":
                accumulable = Accumulable.Max<int>();
                break;
            default:
                throw new NotImplementedException();
        }
        var nextCall = call.Arguments[0] as MethodCallExpression;
        if (nextCall != null)
        {
            accumulable = ConvertNodeToNode(nextCall, accumulable);
        }
        return accumulable;
    }
    private static IAccumulable<int, int> ConvertNodeToNode(
        MethodCallExpression call,
        IAccumulable<int, int> child)
    {
        var accumulable = default(IAccumulable<int, int>);
        switch (call.Method.Name)
        {
            case "Select":
                {
                    var func = call.Arguments[1] as Expression<Func<int, int>>;
                    accumulable = Accumulable.Select(func.Compile(), child);
                    break;
                }
            case "Where":
                {
                    var func = call.Arguments[1] as Expression<Func<int, bool>>;
                    accumulable = Accumulable.Where(func.Compile(), child);
                    break;
                }
            default:
                throw new NotImplementedException();
        }
        var nextCall = call.Arguments[0] as MethodCallExpression;
        if (nextCall != null)
        {
            accumulable = ConvertNodeToNode(nextCall, accumulable);
        }
        return accumulable;
    }
}


static class Program
{
    static void Main(string[] args)
    {
        var q = Enumerable.Range(0, 100)
                  .Hook(new EnumeratorHook<int>()
                  {
                      OnBeforeBegin = () => Console.WriteLine("Beginning!"),
                      OnAfterDispose = _ => Console.WriteLine("Disposed!"),
                  });

        var result = q.ConcurrentReduce(
                          seq => new
                          {
                              Sum = seq.Sum(),
                              Count = seq.Count(),
                              Min = seq.Min(),
                              Max = seq.Max(),
                              X = seq.Where(x => x % 2 == 0)
                                     .Select(x => x * (x - 1))
                                     .Sum(),
                          });

        Console.WriteLine("Result = {0}", result);
    }
}